Require Import Arith.

Require Import Ideal.


Definition set := nat -> bool.
Definition empty : set := fun _ => false.

(** Functional trees, to model heap pieces that represent trees *)
Inductive ftree : Type :=
| Leaf : ftree
| Node : ftree -> nat -> ftree -> ftree.

Module Type TREE.
  Parameter tree : set -> nat -> sprop.
  Parameter ftreeOk : set -> ftree -> nat -> sprop.

  Axiom unfold_tree_fwd : forall s a, tree s a ===> Ex t, !{ftreeOk s t a}.
  Axiom unfold_tree_bwd : forall s a, Ex t, !{ftreeOk s t a} ===> tree s a.

  Axiom ftree_empty_fwd : forall s t a,
    a = 0
    -> ftreeOk s t a ===> [< s = empty >].

  Axiom ftree_nonempty_fwd : forall s t a,
    a <> 0
    -> ftreeOk s t a
    ===> Ex t1, Ex n, Ex t2, Ex s1, Ex a1, Ex s2, Ex a2,
    [< s n = true
      /\ s1 = (fun n' => if le_gt_dec n n' then false else s n')
      /\ s2 = (fun n' => if le_gt_dec n' n then false else s n') >]
    * !{ftreeOk s1 t1 a1}
    * a ==> a1 * (a+1) ==> n * (a+2) ==> a2
    * !{ftreeOk s2 t2 a2}.

  Axiom ftree_nonempty_bwd : forall s t a,
    a <> 0
    -> (Ex t1, Ex n, Ex t2, Ex s1, Ex a1, Ex s2, Ex a2,
    [< t = Node t1 n t2
      /\ s n = true
      /\ s1 = (fun n' => if le_gt_dec n n' then false else s n')
      /\ s2 = (fun n' => if le_gt_dec n' n then false else s n') >]
    * !{ftreeOk s1 t1 a1}
    * a ==> a1 * (a+1) ==> n * (a+2) ==> a2
    * !{ftreeOk s2 t2 a2})
    ===> ftreeOk s t a.
End TREE.

Module Tree : TREE.
  Open Scope Sep_scope.

  Fixpoint ftreeOk (s : set) (t : ftree) (a : nat) : sprop :=
    match t with
      | Leaf => [< a = 0 /\ s = empty >]
      | Node t1 n t2 => Ex s1, Ex a1, Ex s2, Ex a2,
        [< a <> 0
          /\ s n = true
          /\ s1 = (fun n' => if le_gt_dec n n' then false else s n')
          /\ s2 = (fun n' => if le_gt_dec n' n then false else s n') >]
         * !{ftreeOk s1 t1 a1}
         * a ==> a1 * (a+1) ==> n * (a+2) ==> a2
         * !{ftreeOk s2 t2 a2}
    end.

  Definition tree (s : set) (a : nat) : sprop := Ex t, !{ftreeOk s t a}.

  Theorem unfold_tree_fwd : forall s a, tree s a ===> Ex t, !{ftreeOk s t a}.
    sepLemma.
  Qed.

  Theorem unfold_tree_bwd : forall s a, Ex t, !{ftreeOk s t a} ===> tree s a.
    sepLemma.
  Qed.

  Theorem ftree_empty_fwd : forall s t a,
    a = 0
    -> ftreeOk s t a ===> [< s = empty >].
    destruct t; sepLemma.
  Qed.

  Theorem ftree_nonempty_fwd : forall s t a,
    a <> 0
    -> ftreeOk s t a
    ===> Ex t1, Ex n, Ex t2, Ex s1, Ex a1, Ex s2, Ex a2,
    [< s n = true
      /\ s1 = (fun n' => if le_gt_dec n n' then false else s n')
      /\ s2 = (fun n' => if le_gt_dec n' n then false else s n') >]
    * !{ftreeOk s1 t1 a1}
    * a ==> a1 * (a+1) ==> n * (a+2) ==> a2
    * !{ftreeOk s2 t2 a2}.
    destruct t; sepLemma; auto.
  Qed.

  Theorem ftree_nonempty_bwd : forall s t a,
    a <> 0
    -> (Ex t1, Ex n, Ex t2, Ex s1, Ex a1, Ex s2, Ex a2,
      [< t = Node t1 n t2
        /\ s n = true
        /\ s1 = (fun n' => if le_gt_dec n n' then false else s n')
        /\ s2 = (fun n' => if le_gt_dec n' n then false else s n') >]
      * !{ftreeOk s1 t1 a1}
      * a ==> a1 * (a+1) ==> n * (a+2) ==> a2
      * !{ftreeOk s2 t2 a2})
    ===> ftreeOk s t a.
    destruct t; sepLemma;
    try match goal with
          | [ H : @eq ftree _ _ |- _ ] => simplify_eq H; clear H; intros
        end;
    repeat match goal with
             | [ x : _ |- _ ] => subst x
           end; assumption || canceler.
  Qed.
End Tree.

Import Tree.

Hint Immediate unfold_tree_fwd : Forward.
Hint Immediate unfold_tree_bwd : Backward.

Definition treeS : state -> PropX pc state := st ~> Ex fr, Ex s, ![ !{tree s st#R0} * ![fr] ] st
  /\ st#Rret @@ (st' ~> [< st'#R0 = if s st#R1 then 1 else 0 >]
    /\ ![ !{tree s st#R0} * ![fr] ] st').

Definition bst := bmodule {{
  bfunction "lookup" [treeS] {
    [treeS]
    While (R0 != 0) {
      If ($[R0+1] == R1) {
        Use [ftree_nonempty_fwd];;

        R0 <- 1;;
        Goto Rret
      } else {
        If (R1 < $[R0+1]) {
          R0 <- $[R0]
        } else {
          R0 <- $[R0+2]
        }
      };;

      Use [ftree_nonempty_fwd];;
      Use [ftree_nonempty_bwd]
    };;

    Use [ftree_empty_fwd];;

    R0 <- 0;;
    Goto Rret
  }
}}.

Lemma foundIt : forall (s : set) y v,
  y = v
  -> s y = true
  -> 1 = (if s v then 1 else 0).
  intros ? ? ? ? H; subst; rewrite H; auto.
Qed.

Lemma zig_gt : forall x x1 x2 v,
  x1 > x2
  -> x = (if if le_gt_dec x1 x2 then false else v then 1 else 0)
  -> x = (if v then 1 else 0).
  intros; destruct (le_gt_dec x1 x2); auto.
Qed.

Lemma zig_lt : forall x x1 x2 v,
  x1 <= x2
  -> x1 <> x2
  -> x = (if if le_gt_dec x2 x1 then false else v then 1 else 0)
  -> x = (if v then 1 else 0).
  intros; destruct (le_gt_dec x2 x1); auto.
Qed.

Hint Immediate foundIt zig_gt zig_lt.

Theorem bstOk : moduleOk bst.
  structuredSep; eauto.
Qed.
